@@ -5,12 +5,13 @@
 
 use std::borrow::Cow;
 use std::collections::HashMap;
+use std::convert::TryInto;
 
+use anyhow::Context;
 use k8s_openapi::api::core::v1::{ConfigMap, Namespace, Node, Pod, Secret};
 use kube::api::{self, Api};
-// use kube::api::{Api, ListParams, PostParams, Resource, WatchEvent};
-use kube::config::Kubeconfig;
-use kube::{Client, ResourceExt};
+use kube::config::{Config, KubeConfigOptions, Kubeconfig};
+use kube::{Client, Resource, ResourceExt};
 use serde_json as json;
 
 use crate::v0;
@@ -25,53 +26,72 @@ mod helper;
 mod kubeconfig;
 mod show;
 
-const DEFAULT_NS: &str = "default";
-const KUBE_SYSTEM_NS: &str = "kube-system";
+pub(crate) const KUBE_SYSTEM_NS: &str = "kube-system";
 const STATEHUB_CLUSTER_TOKEN_SECRET_TYPE: &str = "statehub.io/cluster-token";
 const STATEHUB_CLUSTER_TOKEN_SECRET_NAME: &str = "statehub-cluster-token";
 const STATEHUB_CLUSTER_CONFIGMAP_NAME: &str = "statehub";
 const STATEHUB_DEFAULT_CLEANUP_GRACE: &str = "600s";
 
 pub(crate) struct Kubectl {
-    client: Client,
-    namespace: String,
+    kubeconfig: Kubeconfig,
+    context: String,
 }
 
 impl Kubectl {
-    async fn with_namespace(namespace: &str) -> anyhow::Result<Self> {
-        let client = Client::try_default().await?;
-        let namespace = namespace.to_string();
-        Ok(Self { client, namespace })
+    pub(crate) async fn new() -> anyhow::Result<Self> {
+        let kubeconfig = Kubeconfig::read()?;
+        let context = kubeconfig
+            .current_context()
+            .ok_or_else(|| anyhow::anyhow!("No current context in kubeconfig"))?
+            .to_string();
+        Ok(Self {
+            kubeconfig,
+            context,
+        })
     }
 
-    pub(crate) async fn default() -> anyhow::Result<Self> {
-        Self::with_namespace(DEFAULT_NS).await
+    pub(crate) async fn with_context(context: &str) -> anyhow::Result<Self> {
+        let kubeconfig = Kubeconfig::read()?;
+        let context = context.to_string();
+
+        Ok(Self {
+            kubeconfig,
+            context,
+        })
     }
 
-    pub(crate) async fn kube_system() -> anyhow::Result<Self> {
-        Self::with_namespace(KUBE_SYSTEM_NS).await
+    pub(crate) fn cluster_name(&self) -> v0::ClusterName {
+        normalize_name(&self.context)
+    }
+
+    async fn get_client(&self) -> anyhow::Result<Client> {
+        self.kubeconfig
+            .context_config(&self.context)
+            .await?
+            .try_into()
+            .context("Connecting to Kubernetes cluster")
     }
 
     async fn all_namespaces(&self) -> anyhow::Result<impl IntoIterator<Item = Namespace>> {
-        let namespaces = self.namespaces();
+        let namespaces = self.api().await?;
         let lp = self.list_params();
         Ok(namespaces.list(&lp).await?)
     }
 
     async fn all_nodes(&self) -> anyhow::Result<impl IntoIterator<Item = Node>> {
-        let nodes = self.nodes();
+        let nodes = self.api().await?;
         let lp = self.list_params();
         Ok(nodes.list(&lp).await?)
     }
 
-    async fn all_pods(&self) -> anyhow::Result<impl IntoIterator<Item = Pod>> {
-        let pods = self.pods();
+    async fn all_pods(&self, namespace: &str) -> anyhow::Result<impl IntoIterator<Item = Pod>> {
+        let pods = self.namespaced_api(namespace).await?;
         let lp = self.list_params();
         Ok(pods.list(&lp).await?)
     }
 
     async fn create_namespace(&self, namespace: &str) -> anyhow::Result<Namespace> {
-        let namespaces = self.namespaces();
+        let namespaces = self.api().await?;
         let namespace = json::from_value(json::json!({
             "apiVerion": "v1",
             "kind": "Namespace",
@@ -87,17 +107,18 @@ impl Kubectl {
     async fn create_configmap(
         &self,
         name: &str,
+        namespace: &str,
         cluster_name: &v0::ClusterName,
         default_state: &str,
         api: &str,
     ) -> anyhow::Result<ConfigMap> {
-        let configmaps = self.configmaps();
+        let configmaps = self.namespaced_api(namespace).await?;
         let configmap = json::from_value(json::json!({
             "apiVerion": "v1",
             "kind": "ConfigMap",
             "metadata": {
                 "name": name,
-                "namespace": self.namespace,
+                "namespace": namespace,
             },
             "data": {
                 "cluster-name": cluster_name,
@@ -114,11 +135,12 @@ impl Kubectl {
 
     async fn create_secret(
         &self,
+        namespace: &str,
         r#type: &str,
         name: &str,
         secret: &str,
     ) -> anyhow::Result<Secret> {
-        let secrets = self.secrets();
+        let secrets = self.namespaced_api(namespace).await?;
         let secret = json::from_value(json::json!({
             "apiVerion": "v1",
             "kind": "Secret",
@@ -135,9 +157,9 @@ impl Kubectl {
         Ok(secret)
     }
 
-    async fn delete_secret(&self, secret: &str) -> anyhow::Result<()> {
+    async fn delete_secret(&self, namespace: &str, secret: &str) -> anyhow::Result<()> {
         log::info!("Deleting secret {}", secret);
-        let secrets = self.secrets();
+        let secrets = self.namespaced_api::<Secret>(namespace).await?;
         let dp = self.delete_params();
         secrets
             .delete(secret, &dp)
@@ -148,9 +170,9 @@ impl Kubectl {
         Ok(())
     }
 
-    async fn delete_configmap(&self, configmap: &str) -> anyhow::Result<()> {
+    async fn delete_configmap(&self, namespace: &str, configmap: &str) -> anyhow::Result<()> {
         log::info!("Deleting configmap {}", configmap);
-        let configmaps = self.configmaps();
+        let configmaps = self.namespaced_api::<ConfigMap>(namespace).await?;
         let dp = self.delete_params();
         configmaps
             .delete(configmap, &dp)
@@ -173,58 +195,95 @@ impl Kubectl {
         api::PostParams::default()
     }
 
-    fn namespaces(&self) -> Api<Namespace> {
-        Api::all(self.client.clone())
+    async fn api<T>(&self) -> anyhow::Result<Api<T>>
+    where
+        T: Resource,
+        <T as Resource>::DynamicType: Default,
+    {
+        self.get_client().await.map(Api::all)
     }
 
-    fn nodes(&self) -> Api<Node> {
-        Api::all(self.client.clone())
+    async fn namespaced_api<T>(&self, namespace: &str) -> anyhow::Result<Api<T>>
+    where
+        T: Resource,
+        <T as Resource>::DynamicType: Default,
+    {
+        self.get_client()
+            .await
+            .map(|client| Api::namespaced(client, namespace))
     }
 
-    fn pods(&self) -> Api<Pod> {
-        Api::namespaced(self.client.clone(), &self.namespace)
+    pub(crate) async fn get_regions(
+        &self,
+        zone: bool,
+    ) -> anyhow::Result<HashMap<Option<String>, Vec<String>>> {
+        let group_nodes = |nodes| {
+            if zone {
+                group_nodes_by_zone(nodes)
+            } else {
+                group_nodes_by_region(nodes)
+            }
+        };
+
+        self.all_nodes().await.map(group_nodes)
+    }
+    pub(crate) async fn collect_node_locations(&self) -> anyhow::Result<Vec<Location>> {
+        self.get_regions(false)
+            .await?
+            .into_iter()
+            .map(|(region, nodes)| {
+                region.ok_or_else(|| {
+                    anyhow::anyhow!("Cannot determine location for nodes {:?}", nodes)
+                })
+            })
+            .collect::<Result<Vec<_>, _>>()?
+            .into_iter()
+            .map(|region| region.parse())
+            .collect::<Result<Vec<Location>, _>>()
+            .map_err(anyhow::Error::msg)
     }
 
-    fn secrets(&self) -> Api<Secret> {
-        Api::namespaced(self.client.clone(), &self.namespace)
-    }
-
-    fn configmaps(&self) -> Api<ConfigMap> {
-        Api::namespaced(self.client.clone(), &self.namespace)
-    }
-}
-
-pub(crate) async fn get_regions(
-    zone: bool,
-) -> anyhow::Result<HashMap<Option<String>, Vec<String>>> {
-    let group_nodes = |nodes| {
-        if zone {
-            group_nodes_by_zone(nodes)
-        } else {
-            group_nodes_by_region(nodes)
+    pub(crate) async fn store_configmap(
+        &self,
+        namespace: &str,
+        cluster_name: &v0::ClusterName,
+        default_state: &str,
+        api: &str,
+    ) -> anyhow::Result<ConfigMap> {
+        if self
+            .delete_configmap(namespace, STATEHUB_CLUSTER_CONFIGMAP_NAME)
+            .await
+            .is_ok()
+        {
+            log::trace!("Removing previous configmap");
         }
-    };
 
-    Kubectl::default().await?.all_nodes().await.map(group_nodes)
-}
+        self.create_configmap(
+            STATEHUB_CLUSTER_CONFIGMAP_NAME,
+            namespace,
+            cluster_name,
+            default_state,
+            api,
+        )
+        .await
+    }
 
-pub(crate) async fn collect_node_locations() -> anyhow::Result<Vec<Location>> {
-    get_regions(false)
-        .await?
-        .into_iter()
-        .map(|(region, nodes)| {
-            region.ok_or_else(|| anyhow::anyhow!("Cannot determine location for nodes {:?}", nodes))
-        })
-        .collect::<Result<Vec<_>, _>>()?
-        .into_iter()
-        .map(|region| region.parse())
-        .collect::<Result<Vec<Location>, _>>()
-        .map_err(anyhow::Error::msg)
+    // TODO: implement provider fetch out of cluster
+    pub(crate) async fn get_cluster_provider(&self) -> anyhow::Result<v0::Provider> {
+        let nodes = self.all_nodes().await?;
+        if is_aks(nodes) {
+            Ok(v0::Provider::Aks)
+        } else if is_eks() {
+            Ok(v0::Provider::Eks)
+        } else {
+            Ok(v0::Provider::Generic)
+        }
+    }
 }
 
 pub(crate) async fn validate_namespace(namespace: impl AsRef<str>) -> anyhow::Result<Namespace> {
     let namespace = namespace.as_ref();
-    let kube = Kubectl::kube_system().await?;
+    let kube = Kubectl::new().await?;
     let existing = kube
         .all_namespaces()
         .await?
@@ -241,59 +300,36 @@ pub(crate) async fn validate_namespace(namespace: impl AsRef<str>) -> anyhow::Re
 }
 
 pub(crate) async fn list_namespaces() -> anyhow::Result<impl IntoIterator<Item = Namespace>> {
-    Kubectl::kube_system().await?.all_namespaces().await
+    Kubectl::new().await?.all_namespaces().await
 }
 
 pub(crate) async fn list_nodes() -> anyhow::Result<impl IntoIterator<Item = Node>> {
-    Kubectl::default().await?.all_nodes().await
+    Kubectl::new().await?.all_nodes().await
 }
 
 pub(crate) async fn list_pods() -> anyhow::Result<impl IntoIterator<Item = Pod>> {
-    Kubectl::kube_system().await?.all_pods().await
-}
-
-pub(crate) async fn store_configmap(
-    namespace: &str,
-    cluster_name: &v0::ClusterName,
-    default_state: &str,
-    api: &str,
-) -> anyhow::Result<ConfigMap> {
-    let kube = Kubectl::with_namespace(namespace).await?;
-
-    if kube
-        .delete_configmap(STATEHUB_CLUSTER_CONFIGMAP_NAME)
-        .await
-        .is_ok()
-    {
-        log::trace!("Removing previous configmap");
-    }
-
-    kube.create_configmap(
-        STATEHUB_CLUSTER_CONFIGMAP_NAME,
-        cluster_name,
-        default_state,
-        api,
-    )
-    .await
+    Kubectl::new().await?.all_pods(KUBE_SYSTEM_NS).await
 }
 
 pub(crate) async fn store_cluster_token(namespace: &str, token: &str) -> anyhow::Result<Secret> {
-    let kube = Kubectl::with_namespace(namespace).await?;
+    let kubectl = Kubectl::new().await?;
 
-    if kube
-        .delete_secret(STATEHUB_CLUSTER_TOKEN_SECRET_NAME)
+    if kubectl
+        .delete_secret(namespace, STATEHUB_CLUSTER_TOKEN_SECRET_NAME)
         .await
         .is_ok()
     {
         log::trace!("Removing previous cluster token");
     }
 
-    kube.create_secret(
-        STATEHUB_CLUSTER_TOKEN_SECRET_TYPE,
-        STATEHUB_CLUSTER_TOKEN_SECRET_NAME,
-        token,
-    )
-    .await
+    kubectl
+        .create_secret(
+            namespace,
+            STATEHUB_CLUSTER_TOKEN_SECRET_TYPE,
+            STATEHUB_CLUSTER_TOKEN_SECRET_NAME,
+            token,
+        )
+        .await
 }
 
 pub(crate) fn helm_is_found() -> bool {
@@ -308,25 +344,16 @@ pub(crate) fn extract_cluster_token(secret: &Secret) -> Option<Cow<'_, str>> {
         .map(String::from_utf8_lossy)
 }
 
-// TODO: implement provider fetch out of cluster
-pub(crate) async fn get_cluster_provider(
-    _cluster: &v0::ClusterName,
-) -> anyhow::Result<v0::Provider> {
-    let kube = Kubectl::default().await?;
-    let nodes = kube.all_nodes().await?;
-    if is_aks(nodes) {
-        Ok(v0::Provider::Aks)
-    } else if is_eks() {
-        Ok(v0::Provider::Eks)
-    } else {
-        Ok(v0::Provider::Generic)
-    }
+pub(crate) fn get_current_cluster_name() -> Option<v0::ClusterName> {
+    Kubeconfig::read().ok()?.current_cluster_name()
 }
 
-pub(crate) fn get_current_cluster_name() -> Option<v0::ClusterName> {
-    Kubeconfig::read()
-        .ok()?
-        .default_context()
-        .map(kubeconfig::normalize_name)
-        .map(v0::ClusterName::from)
+const CLUSTER_DELIMITER: char = '/';
+
+pub(super) fn normalize_name(name: &str) -> v0::ClusterName {
+    if let Some((_, tail)) = name.rsplit_once(CLUSTER_DELIMITER) {
+        tail.into()
+    } else {
+        name.into()
+    }
 }
