@@ -72,8 +72,15 @@ enum Command {
         name: v0::StateName,
         #[structopt(long, short, help = "Defalt owning cluster")]
         owner: Option<v0::ClusterName>,
-        #[structopt(long, short, help = "Location definition")]
-        location: Vec<Location>,
+        #[structopt(long, short, help = "Location definition", alias = "location")]
+        locations: Vec<Location>,
+        #[structopt(
+            long,
+            short,
+            help = "Get locations from this cluster",
+            alias = "cluster"
+        )]
+        clusters: Vec<v0::ClusterName>,
     },
 
     #[structopt(about = "Delete existing state", aliases = &["delete-st", "ds"], display_order(20))]
@@ -388,10 +395,12 @@ impl Cli {
             Command::CreateState {
                 name,
                 owner,
-                location,
+                locations,
+                clusters,
             } => {
-                let locations = location.into();
-                statehub.create_state(name, owner, locations).await
+                statehub
+                    .create_state(name, owner, locations, clusters)
+                    .await
             }
             Command::DeleteState { name: state } => statehub.delete_state(state).await,
             Command::ListStates => statehub.list_states().await,
@@ -568,8 +577,14 @@ impl StateHub {
         &self,
         name: v0::StateName,
         owner: Option<v0::ClusterName>,
-        locations: v0::CreateStateLocationsDto,
+        locations: Vec<Location>,
+        clusters: Vec<v0::ClusterName>,
     ) -> anyhow::Result<()> {
+        let mut locations = vec![locations];
+        for name in &clusters {
+            locations.push(self.api.get_cluster(name).await?.all_locations());
+        }
+        let locations = locations.into_iter().flatten().unique().collect();
         let state = v0::CreateStateDto {
             name,
             storage_class: None,
