@@ -59,9 +59,7 @@ pub enum StatehubError {
         state: StateName,
         volume: VolumeName,
     },
-    UnknownError {
-        message: String,
-    },
+    UnknownError,
 }
 
 #[derive(Clone, Copy, Debug, PartialEq, SerializeDisplay, DeserializeFromStr)]
